#include <iostream>
#include <algorithm>

#include <boost/iterator_adaptors.hpp>

// c_str_iter example:
//   This is a simple iterator adaptor that can be used to pass C-style strings
//   to generic C++ functions taking a pair of iterators as input, without
//   having to determine the length of the C string first.
// Usage:
//   'c_str_iter i(s);' will create an iterator that iterates over the
//     characters in 's' (which is of type 'char *').
//   'c_str_iter i();' will create an iterator that designates an "end" value
//     for any C string.
// In general, [c_str_iter(s), c_str_iter()) is the same as [s, s + strlen(s))
//   EXCEPT that the former is only a ForwardIterator while the latter is a
//   RandomAccessIterator.  However, ForwardIterator's are good enough for many
//   generic C++ functions.
// Constness:
//   const_c_str_iter does the same for 'const char *' as c_str_iter does for
//     'char *'.  const_c_str_iter and c_str_iter will behave properly together
//     as any other const/mutable iterator pair.
// Concepts:
//   A "universal end value" is specific value for an iterator.  An iterator
//     with this value is a "universal end iterator", and may be obtained by
//     using the default constructor.
//   An iterator matches (compares equal to) a univeral end iterator if either
//     of the following is true:
//     . The iterator in question is also a universal end iterator, OR
//     . The iterator in question points to a null character ('\0')
//   If neither iterator in a comparision is a universal end iterator, then the
//     iterators compare equal iff they point to the same character.
//   c_str_iter does not support random access because there is no way to do
//     less-than comparisions with a universal end iterator.
// Future Directions:
//   It is possible that this may be extended into a more generic "sentry value
//     iterator", not limited to C-style strings.  This is left as an exercise
//     for the reader. ;)

// This is an extra iterator wrapper; all it does is provide a default value of
//   0.  It is capable of wrapping any pointer type.
// We just derive directly from iterator_adaptor using the default_iterator_policies,
//   which gives us essentially just a class wrapper for a pointer.
template <typename T>
struct pointer_iterator: public boost::iterator_adaptor<T *, boost::default_iterator_policies>
{
  typedef boost::iterator_adaptor<T *, boost::default_iterator_policies> base_class_type;
  typedef typename base_class_type::base_type base_type;
  typedef typename base_class_type::policies_type policies_type;

  // Note: the 'explicit' on this constructor has been removed, because we want
  //   to be able to create c_str iterators by passing plain pointers.  Since
  //   c_str iterators now take pointer_iterator's in their constructors, the
  //   pointer_iterator constructors must be allowed to be implicit.
  // Alternatively, we could have the c_str iterators derive from iterator_adaptor
  //   and add additional constructors.  This is a better long-term solution, since
  //   it allows the pointer_iterator to be a reusable
  // Note: the default constructor now assigns a well-known value.
  pointer_iterator(const base_type b = 0)
  :base_class_type(b) { }
};

// Inheriting from default_iterator_policies allows us to only specify the
//   policies that change; this is fine with us, since we're not changing how
//   the pointer moves or dereferences -- only how it compares with other pointers.
struct c_str_policy: public boost::default_iterator_policies
{
  // Our test for equality is a little complicated, since we must handle the
  //   cases where either, both, or neither of the iterators is the universal
  //   "end" value.
  // For each iterator, the wrapped pointer_iterator is at ".base()", and
  //   therefore the wrapped pointer is at ".base().base()"
  template <typename IteratorAdaptor1, typename IteratorAdaptor2>
  static bool equal(const IteratorAdaptor1 & x, const IteratorAdaptor2 & y)
  {
    if (y.base().base() == 0)
      return (x.base().base() == 0 || *x.base().base() == '\0');
    if (x.base().base() == 0)
      return (*y.base().base() == '\0');
    return (x.base().base() == y.base().base());
  }
};

// Use the iterator_adaptor class to define our c_str iterators.
typedef boost::iterator_adaptor<pointer_iterator<char>, c_str_policy,
    boost::iterator_category_is<std::forward_iterator_tag> > c_str_iter;
typedef boost::iterator_adaptor<pointer_iterator<const char>, c_str_policy,
    boost::iterator_category_is<std::forward_iterator_tag> > const_c_str_iter;

int main()
{
  // Null-terminated C-style string
  char * const hello_world = "Hello, world!";

  // Copy it to std::cout using a generic C++ function, without accessing the
  //   entire string twice.
  std::cout << "`";
  std::copy(c_str_iter(hello_world), c_str_iter(), std::ostream_iterator<char>(std::cout));
  std::cout << "'\n";

  return 0;
}